#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
pub struct Settings {
    /// Pushes tools' bin-paths to the front of PATH instead of allowing modifications of PATH after activation to take precedence.
    #[config(env = "MISE_ACTIVATE_AGGRESSIVE", default = false)]
    pub activate_aggressive: bool,
    /// do not use precompiled binaries for any tool
    #[config(env = "MISE_ALL_COMPILE", default = false)]
    pub all_compile: bool,
    /// should mise keep downloaded files after installation
    #[config(env = "MISE_ALWAYS_KEEP_DOWNLOAD", default = false)]
    pub always_keep_download: bool,
    /// should mise keep install files after installation even if the installation fails
    #[config(env = "MISE_ALWAYS_KEEP_INSTALL", default = false)]
    pub always_keep_install: bool,
    #[config(nested)]
    pub aqua: SettingsAqua,
    /// Architecture to use for precompiled binaries.
    #[config(env = "MISE_ARCH")]
    pub arch: Option<String>,
    /// use asdf as a default plugin backend
    #[config(env = "MISE_ASDF")]
    pub asdf: Option<bool>,
    /// set to true to ensure .tool-versions will be compatible with asdf
    #[config(env = "MISE_ASDF_COMPAT", default = false)]
    pub asdf_compat: bool,
    /// Automatically install missing tools when running `mise x`, `mise run`, or as part of the 'not found' handler.
    #[config(env = "MISE_AUTO_INSTALL", default = true)]
    pub auto_install: bool,
    /// List of tools to skip automatically installing when running `mise x`, `mise run`, or as part of the 'not found' handler.
    #[config(env = "MISE_AUTO_INSTALL_DISABLE_TOOLS")]
    pub auto_install_disable_tools: Option<Vec<String>>,
    /// Delete files in cache that have not been accessed in this duration
    #[config(env = "MISE_CACHE_PRUNE_AGE", default = "30d")]
    pub cache_prune_age: String,
    #[config(nested)]
    pub cargo: SettingsCargo,
    /// Use cargo-binstall instead of cargo install if available
    #[config()]
    pub cargo_binstall: Option<bool>,
    /// Path to change to after launching mise
    #[config(env = "MISE_CD")]
    pub cd: Option<PathBuf>,
    /// Set to true if running in a CI environment
    #[config(env = "CI", default = "false", deserialize_with = bool_string)]
    pub ci: bool,
    /// Use color in mise terminal output
    #[config(env = "MISE_COLOR", default = true)]
    pub color: bool,
    /// Sets log level to debug
    #[config(env = "MISE_DEBUG", default = false)]
    pub debug: bool,
    /// The default config filename read. `mise use` and other commands that create new config files will use this value. This must be an env var.
    #[config(env = "MISE_DEFAULT_CONFIG_FILENAME", default = "mise.toml")]
    pub default_config_filename: String,
    /// The default .tool-versions filename read. This will not ignore .tool-versions—use override_tool_versions_filename for that. This must be an env var.
    #[config(env = "MISE_DEFAULT_TOOL_VERSIONS_FILENAME", default = ".tool-versions")]
    pub default_tool_versions_filename: String,
    /// Backends to disable such as `asdf` or `pipx`
    #[config(env = "MISE_DISABLE_BACKENDS", default = [], parse_env = list_by_comma)]
    pub disable_backends: Vec<String>,
    /// Disable the default mapping of short tool names like `go` -> `vfox:version-fox/vfox-golang`. This parameter disables only for the backends `vfox` and `asdf`.
    #[config(env = "MISE_DISABLE_DEFAULT_REGISTRY", default = false)]
    pub disable_default_registry: bool,
    /// Disables built-in shorthands to asdf/vfox plugins
    #[config(env = "MISE_DISABLE_DEFAULT_SHORTHANDS")]
    pub disable_default_shorthands: Option<bool>,
    /// Turns off helpful hints when using different mise features
    #[config(env = "MISE_DISABLE_HINTS", default = [], parse_env = set_by_comma)]
    pub disable_hints: BTreeSet<String>,
    /// Tools defined in mise.toml that should be ignored
    #[config(env = "MISE_DISABLE_TOOLS", default = [], parse_env = set_by_comma)]
    pub disable_tools: BTreeSet<String>,
    #[config(nested)]
    pub dotnet: SettingsDotnet,
    /// Tools defined in mise.toml that should be used - all other tools are ignored
    #[config(env = "MISE_ENABLE_TOOLS", default = [], parse_env = set_by_comma)]
    pub enable_tools: BTreeSet<String>,
    /// Env to use for mise.<MISE_ENV>.toml files.
    #[config(env = "MISE_ENV", default = [], parse_env = list_by_comma)]
    pub env: Vec<String>,
    /// Path to a file containing environment variables to automatically load.
    #[config(env = "MISE_ENV_FILE")]
    pub env_file: Option<PathBuf>,
    #[config(nested)]
    pub erlang: SettingsErlang,
    /// Automatically install missing tools when running `mise x`.
    #[config(env = "MISE_EXEC_AUTO_INSTALL", default = true)]
    pub exec_auto_install: bool,
    /// Enable experimental mise features which are incomplete or unstable—breakings changes may occur
    #[config(env = "MISE_EXPERIMENTAL", default = false)]
    pub experimental: bool,
    /// How long to cache remote versions for tools.
    #[config(env = "MISE_FETCH_REMOTE_VERSIONS_CACHE", default = "1h")]
    pub fetch_remote_versions_cache: String,
    /// Timeout in seconds for HTTP requests to fetch new tool versions in mise.
    #[config(env = "MISE_FETCH_REMOTE_VERSIONS_TIMEOUT", default = "5s")]
    pub fetch_remote_versions_timeout: String,
    /// Use gix for git operations, set to false to shell out to git.
    #[config(env = "MISE_GIX", default = true)]
    pub gix: bool,
    /// Path to the global mise config file. Default is `~/.config/mise/config.toml`. This must be an env var.
    #[config(env = "MISE_GLOBAL_CONFIG_FILE")]
    pub global_config_file: Option<PathBuf>,
    /// Path which is used as `{{config_root}}` for the global config file. Default is `$HOME`. This must be an env var.
    #[config(env = "MISE_GLOBAL_CONFIG_ROOT")]
    pub global_config_root: Option<PathBuf>,
    /// Path to a file containing default go packages to install when installing go
    #[config(env = "MISE_GO_DEFAULT_PACKAGES_FILE", default = "~/.default-go-packages")]
    pub go_default_packages_file: PathBuf,
    /// Mirror to download go sdk tarballs from.
    #[config(env = "MISE_GO_DOWNLOAD_MIRROR", default = "https://dl.google.com/go")]
    pub go_download_mirror: String,
    /// URL to fetch go from.
    #[config(env = "MISE_GO_REPO", default = "https://github.com/golang/go")]
    pub go_repo: String,
    /// Changes where `go install` installs binaries to.
    #[config(env = "MISE_GO_SET_GOBIN")]
    pub go_set_gobin: Option<bool>,
    /// [deprecated] Set to true to set GOPATH=~/.local/share/mise/installs/go/.../packages.
    #[config(env = "MISE_GO_SET_GOPATH", default = false)]
    pub go_set_gopath: bool,
    /// Sets GOROOT=~/.local/share/mise/installs/go/.../.
    #[config(env = "MISE_GO_SET_GOROOT", default = true)]
    pub go_set_goroot: bool,
    /// Set to true to skip checksum verification when downloading go sdk tarballs.
    #[config(env = "MISE_GO_SKIP_CHECKSUM", default = false)]
    pub go_skip_checksum: bool,
    /// Use gpg to verify all tool signatures.
    #[config(env = "MISE_GPG_VERIFY")]
    pub gpg_verify: Option<bool>,
    /// Timeout in seconds for all HTTP requests in mise.
    #[config(env = "MISE_HTTP_TIMEOUT", default = "30s")]
    pub http_timeout: String,
    /// Set to false to disable the idiomatic version files such as .node-version, .ruby-version, etc.
    #[config(env = "MISE_IDIOMATIC_VERSION_FILE", default = true)]
    pub idiomatic_version_file: bool,
    /// Specific tools to disable idiomatic version files for.
    #[config(env = "MISE_IDIOMATIC_VERSION_FILE_DISABLE_TOOLS", default = [], parse_env = set_by_comma)]
    pub idiomatic_version_file_disable_tools: BTreeSet<String>,
    /// Specific tools to enable idiomatic version files for like .node-version, .ruby-version, etc.
    #[config(env = "MISE_IDIOMATIC_VERSION_FILE_ENABLE_TOOLS", parse_env = set_by_comma)]
    pub idiomatic_version_file_enable_tools: Option<BTreeSet<String>>,
    /// This is a list of config paths that mise will ignore.
    #[config(env = "MISE_IGNORED_CONFIG_PATHS", default = [], parse_env = list_by_colon)]
    pub ignored_config_paths: BTreeSet<PathBuf>,
    /// How many jobs to run concurrently such as tool installs.
    #[config(env = "MISE_JOBS", default = 8)]
    pub jobs: usize,
    /// Set to false to disable the idiomatic version files such as .node-version, .ruby-version, etc.
    #[config(env = "MISE_LEGACY_VERSION_FILE", default = true)]
    pub legacy_version_file: bool,
    /// Specific tools to disable idiomatic version files for.
    #[config(env = "MISE_LEGACY_VERSION_FILE_DISABLE_TOOLS", default = [], parse_env = set_by_comma)]
    pub legacy_version_file_disable_tools: BTreeSet<String>,
    /// Use libgit2 for git operations, set to false to shell out to git.
    #[config(env = "MISE_LIBGIT2", default = true)]
    pub libgit2: bool,
    /// Create and read lockfiles for tool versions.
    #[config(env = "MISE_LOCKFILE", default = true)]
    pub lockfile: bool,
    /// Show more/less output.
    #[config(env = "MISE_LOG_LEVEL", default = "info")]
    pub log_level: String,
    #[config(nested)]
    pub node: SettingsNode,
    /// Set to false to disable the "command not found" handler to autoinstall missing tool versions.
    #[config(env = "MISE_NOT_FOUND_AUTO_INSTALL", default = true)]
    pub not_found_auto_install: bool,
    #[config(nested)]
    pub npm: SettingsNpm,
    /// OS to use for precompiled binaries.
    #[config(env = "MISE_OS")]
    pub os: Option<String>,
    /// If set, mise will ignore default config files like `mise.toml` and use these filenames instead. This must be an env var.
    #[config(env = "MISE_OVERRIDE_CONFIG_FILENAMES", default = [], parse_env = list_by_colon)]
    pub override_config_filenames: Vec<String>,
    /// If set, mise will ignore .tool-versions files and use these filenames instead. Can be set to `none` to disable .tool-versions. This must be an env var.
    #[config(env = "MISE_OVERRIDE_TOOL_VERSIONS_FILENAMES", default = [], parse_env = list_by_colon)]
    pub override_tool_versions_filenames: Vec<String>,
    /// Enables extra-secure behavior.
    #[config(env = "MISE_PARANOID", default = false)]
    pub paranoid: bool,
    /// Default to pinning versions when running `mise use` in mise.toml files.
    #[config(env = "MISE_PIN", default = false)]
    pub pin: bool,
    #[config(nested)]
    pub pipx: SettingsPipx,
    /// Use uvx instead of pipx if uv is installed and on PATH.
    #[config()]
    pub pipx_uvx: Option<bool>,
    /// How long to wait before updating plugins automatically (note this isn't currently implemented).
    #[config(env = "MISE_PLUGIN_AUTOUPDATE_LAST_CHECK_DURATION", default = "7d")]
    pub plugin_autoupdate_last_check_duration: String,
    /// Profile to use for mise.${MISE_PROFILE}.toml files.
    #[config(env = "MISE_PROFILE")]
    pub profile: Option<String>,
    #[config(nested)]
    pub python: SettingsPython,
    /// If true, compile python from source. If false, use precompiled binaries. If not set, use precompiled binaries if available.
    #[config()]
    pub python_compile: Option<bool>,
    /// Path to a file containing default python packages to install when installing python.
    #[config()]
    pub python_default_packages_file: Option<PathBuf>,
    /// URL to fetch python patches from.
    #[config()]
    pub python_patch_url: Option<String>,
    /// Directory to fetch python patches from.
    #[config()]
    pub python_patches_directory: Option<PathBuf>,
    /// Specify the architecture to use for precompiled binaries.
    #[config()]
    pub python_precompiled_arch: Option<String>,
    /// Specify the OS to use for precompiled binaries.
    #[config()]
    pub python_precompiled_os: Option<String>,
    /// URL to fetch pyenv from for compiling python.
    #[config()]
    pub python_pyenv_repo: Option<String>,
    /// Automatically create virtualenvs for python tools.
    #[config()]
    pub python_venv_auto_create: Option<bool>,
    /// Prefer to use venv from Python's standard library.
    #[config()]
    pub python_venv_stdlib: Option<bool>,
    /// Suppress all output except errors.
    #[config(env = "MISE_QUIET", default = false)]
    pub quiet: bool,
    /// Connect stdin/stdout/stderr to child processes.
    #[config(env = "MISE_RAW", default = false)]
    pub raw: bool,
    #[config(nested)]
    pub ruby: SettingsRuby,
    #[config(nested)]
    pub rust: SettingsRust,
    /// Path to a file containing custom tool shorthands.
    #[config(env = "MISE_SHORTHANDS_FILE")]
    pub shorthands_file: Option<PathBuf>,
    /// Suppress all `mise run|watch` output except errors—including what tasks output.
    #[config(env = "MISE_SILENT", default = false)]
    pub silent: bool,
    #[config(nested)]
    pub sops: SettingsSops,
    #[config(nested)]
    pub status: SettingsStatus,
    #[config(nested)]
    pub swift: SettingsSwift,
    /// Path to the system mise config file. Default is `/etc/mise/config.toml`. This must be an env var.
    #[config(env = "MISE_SYSTEM_CONFIG_FILE")]
    pub system_config_file: Option<PathBuf>,
    /// Paths that mise will not look for tasks in.
    #[config(env = "MISE_TASK_DISABLE_PATHS", default = [], parse_env = list_by_colon)]
    pub task_disable_paths: BTreeSet<PathBuf>,
    /// Change output style when executing tasks.
    #[config(env = "MISE_TASK_OUTPUT")]
    pub task_output: Option<crate::cli::run::TaskOutput>,
    /// Mise will always fetch the latest tasks from the remote, by default the cache is used.
    #[config(env = "MISE_TASK_REMOTE_NO_CACHE")]
    pub task_remote_no_cache: Option<bool>,
    /// Automatically install missing tools when executing tasks.
    #[config(env = "MISE_TASK_RUN_AUTO_INSTALL", default = true)]
    pub task_run_auto_install: bool,
    /// Tasks to skip when running `mise run`.
    #[config(env = "MISE_TASK_SKIP", default = [], parse_env = set_by_comma)]
    pub task_skip: BTreeSet<String>,
    /// Show completion message with elapsed time for each task on `mise run`. Default shows when output type is `prefix`.
    #[config(env = "MISE_TASK_TIMINGS")]
    pub task_timings: Option<bool>,
    /// Sets log level to trace
    #[config(env = "MISE_TRACE", default = false)]
    pub trace: bool,
    /// This is a list of config paths that mise will automatically mark as trusted.
    #[config(env = "MISE_TRUSTED_CONFIG_PATHS", default = [], parse_env = list_by_colon)]
    pub trusted_config_paths: BTreeSet<PathBuf>,
    /// List of default shell arguments for unix to be used with `file`. For example `sh`.
    #[config(env = "MISE_UNIX_DEFAULT_FILE_SHELL_ARGS", default = "sh")]
    pub unix_default_file_shell_args: String,
    /// List of default shell arguments for unix to be used with inline commands. For example, `sh`, `-c` for sh.
    #[config(env = "MISE_UNIX_DEFAULT_INLINE_SHELL_ARGS", default = "sh -c -o errexit")]
    pub unix_default_inline_shell_args: String,
    /// Determines whether to use a specified shell for executing tasks in the tasks directory. When set to true, the shell defined in the file will be used, or the default shell specified by `windows_default_file_shell_args` or `unix_default_file_shell_args` will be applied. If set to false, tasks will be executed directly as programs.
    #[config(env = "MISE_USE_FILE_SHELL_FOR_EXECUTABLE_TASKS", default = false)]
    pub use_file_shell_for_executable_tasks: bool,
    /// Set to false to disable using the mise-versions API as a quick way for mise to query for new versions.
    #[config(env = "MISE_USE_VERSIONS_HOST", default = true)]
    pub use_versions_host: bool,
    /// Shows more verbose output such as installation logs when installing tools.
    #[config(env = "MISE_VERBOSE", default = false)]
    pub verbose: bool,
    /// Use vfox as a default plugin backend instead of asdf.
    #[config(env = "MISE_VFOX")]
    pub vfox: Option<bool>,
    /// List of default shell arguments for Windows to be used for file commands. For example, `cmd`, `/c` for cmd.exe.
    #[config(env = "MISE_WINDOWS_DEFAULT_FILE_SHELL_ARGS", default = "cmd /c")]
    pub windows_default_file_shell_args: String,
    /// List of default shell arguments for Windows to be used for inline commands. For example, `cmd`, `/c` for cmd.exe.
    #[config(env = "MISE_WINDOWS_DEFAULT_INLINE_SHELL_ARGS", default = "cmd /c")]
    pub windows_default_inline_shell_args: String,
    /// List of executable extensions for Windows. For example, `exe` for .exe files, `bat` for .bat files, and so on.
    #[config(env = "MISE_WINDOWS_EXECUTABLE_EXTENSIONS", default = ["exe", "bat", "cmd", "com", "ps1", "vbs"], parse_env = list_by_comma)]
    pub windows_executable_extensions: Vec<String>,
    /// Shim file mode for Windows. Options: `file`, `hardlink`, `symlink`.
    #[config(env = "MISE_WINDOWS_SHIM_MODE", default = "file")]
    pub windows_shim_mode: String,
    /// This will automatically answer yes or no to prompts. This is useful for scripting.
    #[config(env = "MISE_YES", default = false)]
    pub yes: bool,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsAqua {
    /// Use baked-in aqua registry (if compiled in).
    #[config(env = "MISE_AQUA_BAKED_REGISTRY", default = true)]
    pub baked_registry: bool,
    /// Use cosign to verify aqua tool signatures.
    #[config(env = "MISE_AQUA_COSIGN", default = true)]
    pub cosign: bool,
    /// Extra arguments to pass to cosign when verifying aqua tool signatures.
    #[config(env = "MISE_AQUA_COSIGN_EXTRA_ARGS")]
    pub cosign_extra_args: Option<Vec<String>>,
    /// Use minisign to verify aqua tool signatures.
    #[config(env = "MISE_AQUA_MINISIGN", default = true)]
    pub minisign: bool,
    /// URL to fetch aqua registry from.
    #[config(env = "MISE_AQUA_REGISTRY_URL")]
    pub registry_url: Option<String>,
    /// Use SLSA to verify aqua tool signatures.
    #[config(env = "MISE_AQUA_SLSA", default = true)]
    pub slsa: bool,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsCargo {
    /// Use cargo-binstall instead of cargo install if available
    #[config(env = "MISE_CARGO_BINSTALL", default = true)]
    pub binstall: bool,
    /// Name of the cargo registry to use.
    #[config(env = "MISE_CARGO_REGISTRY_NAME")]
    pub registry_name: Option<String>,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsDotnet {
    /// Extends dotnet search and install abilities.
    #[config(env = "MISE_DOTNET_PACKAGE_FLAGS", default = [], parse_env = list_by_comma)]
    pub package_flags: Vec<String>,
    /// URL to fetch dotnet tools from.
    #[config(env = "MISE_DOTNET_REGISTRY_URL", default = "https://api.nuget.org/v3/index.json")]
    pub registry_url: String,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsErlang {
    /// If true, compile erlang from source. If false, use precompiled binaries. If not set, use precompiled binaries if available.
    #[config(env = "MISE_ERLANG_COMPILE")]
    pub compile: Option<bool>,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsNode {
    /// Compile node from source.
    #[config(env = "MISE_NODE_COMPILE")]
    pub compile: Option<bool>,
    /// Install a specific node flavor like glibc-217 or musl. Use with unofficial node build repo.
    #[config(env = "MISE_NODE_FLAVOR")]
    pub flavor: Option<String>,
    /// Use gpg to verify node tool signatures.
    #[config(env = "MISE_NODE_GPG_VERIFY")]
    pub gpg_verify: Option<bool>,
    /// Mirror to download node tarballs from.
    #[config(env = "MISE_NODE_MIRROR_URL")]
    pub mirror_url: Option<String>,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsNpm {
    /// Use bun instead of npm if bun is installed and on PATH.
    #[config(env = "MISE_NPM_BUN", default = false)]
    pub bun: bool,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsPipx {
    /// URL to use for pipx registry.
    #[config(env = "MISE_PIPX_REGISTRY_URL", default = "https://pypi.org/pypi/{}/json")]
    pub registry_url: String,
    /// Use uvx instead of pipx if uv is installed and on PATH.
    #[config(env = "MISE_PIPX_UVX")]
    pub uvx: Option<bool>,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsPython {
    /// If true, compile python from source. If false, use precompiled binaries. If not set, use precompiled binaries if available.
    #[config(env = "MISE_PYTHON_COMPILE")]
    pub compile: Option<bool>,
    /// Path to a file containing default python packages to install when installing a python version.
    #[config(env = "MISE_PYTHON_DEFAULT_PACKAGES_FILE")]
    pub default_packages_file: Option<PathBuf>,
    /// URL to fetch python patches from to pass to python-build.
    #[config(env = "MISE_PYTHON_PATCH_URL")]
    pub patch_url: Option<String>,
    /// Directory to fetch python patches from.
    #[config(env = "MISE_PYTHON_PATCHES_DIRECTORY")]
    pub patches_directory: Option<PathBuf>,
    /// Specify the architecture to use for precompiled binaries.
    #[config(env = "MISE_PYTHON_PRECOMPILED_ARCH")]
    pub precompiled_arch: Option<String>,
    /// Specify the flavor to use for precompiled binaries.
    #[config(env = "MISE_PYTHON_PRECOMPILED_FLAVOR")]
    pub precompiled_flavor: Option<String>,
    /// Specify the OS to use for precompiled binaries.
    #[config(env = "MISE_PYTHON_PRECOMPILED_OS")]
    pub precompiled_os: Option<String>,
    /// URL to fetch pyenv from for compiling python with python-build.
    #[config(env = "MISE_PYENV_REPO", default = "https://github.com/pyenv/pyenv.git")]
    pub pyenv_repo: String,
    /// Integrate with uv to automatically create/source venvs if uv.lock is present.
    #[config(env = "MISE_PYTHON_UV_VENV_AUTO", default = false)]
    pub uv_venv_auto: bool,
    /// Arguments to pass to uv when creating a venv.
    #[config(env = "MISE_PYTHON_UV_VENV_CREATE_ARGS", parse_env = list_by_colon)]
    pub uv_venv_create_args: Option<Vec<String>>,
    /// Automatically create virtualenvs for python tools.
    #[config(env = "MISE_PYTHON_VENV_AUTO_CREATE", default = false)]
    pub venv_auto_create: bool,
    /// Arguments to pass to python when creating a venv. (not used for uv venv creation)
    #[config(env = "MISE_PYTHON_VENV_CREATE_ARGS", parse_env = list_by_colon)]
    pub venv_create_args: Option<Vec<String>>,
    /// Prefer to use venv from Python's standard library.
    #[config(env = "MISE_VENV_STDLIB", default = false)]
    pub venv_stdlib: bool,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsRuby {
    /// A list of patch files or URLs to apply to ruby source.
    #[config(env = "MISE_RUBY_APPLY_PATCHES")]
    pub apply_patches: Option<String>,
    /// Path to a file containing default ruby gems to install when installing ruby.
    #[config(env = "MISE_RUBY_DEFAULT_PACKAGES_FILE", default = "~/.default-gems")]
    pub default_packages_file: String,
    /// Options to pass to ruby-build.
    #[config(env = "MISE_RUBY_BUILD_OPTS")]
    pub ruby_build_opts: Option<String>,
    /// URL to fetch ruby-build from.
    #[config(env = "MISE_RUBY_BUILD_REPO", default = "https://github.com/rbenv/ruby-build.git")]
    pub ruby_build_repo: String,
    /// Use ruby-install instead of ruby-build.
    #[config(env = "MISE_RUBY_INSTALL", default = false)]
    pub ruby_install: bool,
    /// Options to pass to ruby-install.
    #[config(env = "MISE_RUBY_INSTALL_OPTS")]
    pub ruby_install_opts: Option<String>,
    /// URL to fetch ruby-install from.
    #[config(env = "MISE_RUBY_INSTALL_REPO", default = "https://github.com/postmodern/ruby-install.git")]
    pub ruby_install_repo: String,
    /// Set to true to enable verbose output during ruby installation.
    #[config(env = "MISE_RUBY_VERBOSE_INSTALL")]
    pub verbose_install: Option<bool>,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsRust {
    /// Path to the cargo home directory. Defaults to `~/.cargo` or `%USERPROFILE%\.cargo`
    #[config(env = "MISE_CARGO_HOME")]
    pub cargo_home: Option<PathBuf>,
    /// Path to the rustup home directory. Defaults to `~/.rustup` or `%USERPROFILE%\.rustup`
    #[config(env = "MISE_RUSTUP_HOME")]
    pub rustup_home: Option<PathBuf>,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsSops {
    /// The age private key to use for sops secret decryption.
    #[config(env = "MISE_SOPS_AGE_KEY")]
    pub age_key: Option<String>,
    /// Path to the age private key file to use for sops secret decryption.
    #[config(env = "MISE_SOPS_AGE_KEY_FILE")]
    pub age_key_file: Option<PathBuf>,
    /// The age public keys to use for sops secret encryption.
    #[config(env = "MISE_SOPS_AGE_RECIPIENTS")]
    pub age_recipients: Option<String>,
    /// Use rops to decrypt sops files. Disable to shell out to `sops` which will slow down mise but sops may offer features not available in rops.
    #[config(env = "MISE_SOPS_ROPS", default = true)]
    pub rops: bool,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsStatus {
    /// Show a warning if tools are not installed when entering a directory with a mise.toml file.
    #[config(env = "MISE_STATUS_MESSAGE_MISSING_TOOLS", default = "if_other_versions_installed")]
    pub missing_tools: String,
    /// Show configured env vars when entering a directory with a mise.toml file.
    #[config(env = "MISE_STATUS_MESSAGE_SHOW_ENV", default = false)]
    pub show_env: bool,
    /// Show configured tools when entering a directory with a mise.toml file.
    #[config(env = "MISE_STATUS_MESSAGE_SHOW_TOOLS", default = false)]
    pub show_tools: bool,
    /// Truncate status messages.
    #[config(env = "MISE_STATUS_MESSAGE_TRUNCATE", default = true)]
    pub truncate: bool,
}

#[derive(Config, Default, Debug, Clone, Serialize)]
#[config(partial_attr(derive(Clone, Serialize, Default)))]
#[config(partial_attr(serde(deny_unknown_fields)))]
pub struct SettingsSwift {
    /// Use gpg to verify swift tool signatures.
    #[config(env = "MISE_SWIFT_GPG_VERIFY")]
    pub gpg_verify: Option<bool>,
    /// Override the platform to use for precompiled binaries.
    #[config(env = "MISE_SWIFT_PLATFORM")]
    pub platform: Option<String>,
}

pub static SETTINGS_META: Lazy<IndexMap<&'static str, SettingsMeta>> = Lazy::new(|| {
    indexmap!{
    "activate_aggressive" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Pushes tools' bin-paths to the front of PATH instead of allowing modifications of PATH after activation to take precedence."###,
    },
    "all_compile" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"do not use precompiled binaries for any tool"###,
    },
    "always_keep_download" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"should mise keep downloaded files after installation"###,
    },
    "always_keep_install" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"should mise keep install files after installation even if the installation fails"###,
    },
    "arch" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Architecture to use for precompiled binaries."###,
    },
    "asdf" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"use asdf as a default plugin backend"###,
    },
    "asdf_compat" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"set to true to ensure .tool-versions will be compatible with asdf"###,
    },
    "auto_install" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Automatically install missing tools when running `mise x`, `mise run`, or as part of the 'not found' handler."###,
    },
    "auto_install_disable_tools" => SettingsMeta {
        type_: SettingsType::ListString,
        description: r###"List of tools to skip automatically installing when running `mise x`, `mise run`, or as part of the 'not found' handler."###,
    },
    "cache_prune_age" => SettingsMeta {
        type_: SettingsType::Duration,
        description: r###"Delete files in cache that have not been accessed in this duration"###,
    },
    "cargo_binstall" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use cargo-binstall instead of cargo install if available"###,
    },
    "cd" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Path to change to after launching mise"###,
    },
    "ci" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Set to true if running in a CI environment"###,
    },
    "color" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use color in mise terminal output"###,
    },
    "debug" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Sets log level to debug"###,
    },
    "default_config_filename" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"The default config filename read. `mise use` and other commands that create new config files will use this value. This must be an env var."###,
    },
    "default_tool_versions_filename" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"The default .tool-versions filename read. This will not ignore .tool-versions—use override_tool_versions_filename for that. This must be an env var."###,
    },
    "disable_backends" => SettingsMeta {
        type_: SettingsType::ListString,
        description: r###"Backends to disable such as `asdf` or `pipx`"###,
    },
    "disable_default_registry" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Disable the default mapping of short tool names like `go` -> `vfox:version-fox/vfox-golang`. This parameter disables only for the backends `vfox` and `asdf`."###,
    },
    "disable_default_shorthands" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Disables built-in shorthands to asdf/vfox plugins"###,
    },
    "disable_hints" => SettingsMeta {
        type_: SettingsType::SetString,
        description: r###"Turns off helpful hints when using different mise features"###,
    },
    "disable_tools" => SettingsMeta {
        type_: SettingsType::SetString,
        description: r###"Tools defined in mise.toml that should be ignored"###,
    },
    "enable_tools" => SettingsMeta {
        type_: SettingsType::SetString,
        description: r###"Tools defined in mise.toml that should be used - all other tools are ignored"###,
    },
    "env" => SettingsMeta {
        type_: SettingsType::ListString,
        description: r###"Env to use for mise.<MISE_ENV>.toml files."###,
    },
    "env_file" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Path to a file containing environment variables to automatically load."###,
    },
    "exec_auto_install" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Automatically install missing tools when running `mise x`."###,
    },
    "experimental" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Enable experimental mise features which are incomplete or unstable—breakings changes may occur"###,
    },
    "fetch_remote_versions_cache" => SettingsMeta {
        type_: SettingsType::Duration,
        description: r###"How long to cache remote versions for tools."###,
    },
    "fetch_remote_versions_timeout" => SettingsMeta {
        type_: SettingsType::Duration,
        description: r###"Timeout in seconds for HTTP requests to fetch new tool versions in mise."###,
    },
    "gix" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use gix for git operations, set to false to shell out to git."###,
    },
    "global_config_file" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Path to the global mise config file. Default is `~/.config/mise/config.toml`. This must be an env var."###,
    },
    "global_config_root" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Path which is used as `{{config_root}}` for the global config file. Default is `$HOME`. This must be an env var."###,
    },
    "go_default_packages_file" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Path to a file containing default go packages to install when installing go"###,
    },
    "go_download_mirror" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Mirror to download go sdk tarballs from."###,
    },
    "go_repo" => SettingsMeta {
        type_: SettingsType::Url,
        description: r###"URL to fetch go from."###,
    },
    "go_set_gobin" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Changes where `go install` installs binaries to."###,
    },
    "go_set_gopath" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"[deprecated] Set to true to set GOPATH=~/.local/share/mise/installs/go/.../packages."###,
    },
    "go_set_goroot" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Sets GOROOT=~/.local/share/mise/installs/go/.../."###,
    },
    "go_skip_checksum" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Set to true to skip checksum verification when downloading go sdk tarballs."###,
    },
    "gpg_verify" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use gpg to verify all tool signatures."###,
    },
    "http_timeout" => SettingsMeta {
        type_: SettingsType::Duration,
        description: r###"Timeout in seconds for all HTTP requests in mise."###,
    },
    "idiomatic_version_file" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Set to false to disable the idiomatic version files such as .node-version, .ruby-version, etc."###,
    },
    "idiomatic_version_file_disable_tools" => SettingsMeta {
        type_: SettingsType::SetString,
        description: r###"Specific tools to disable idiomatic version files for."###,
    },
    "idiomatic_version_file_enable_tools" => SettingsMeta {
        type_: SettingsType::SetString,
        description: r###"Specific tools to enable idiomatic version files for like .node-version, .ruby-version, etc."###,
    },
    "ignored_config_paths" => SettingsMeta {
        type_: SettingsType::ListPath,
        description: r###"This is a list of config paths that mise will ignore."###,
    },
    "jobs" => SettingsMeta {
        type_: SettingsType::Integer,
        description: r###"How many jobs to run concurrently such as tool installs."###,
    },
    "legacy_version_file" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Set to false to disable the idiomatic version files such as .node-version, .ruby-version, etc."###,
    },
    "legacy_version_file_disable_tools" => SettingsMeta {
        type_: SettingsType::SetString,
        description: r###"Specific tools to disable idiomatic version files for."###,
    },
    "libgit2" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use libgit2 for git operations, set to false to shell out to git."###,
    },
    "lockfile" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Create and read lockfiles for tool versions."###,
    },
    "log_level" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Show more/less output."###,
    },
    "not_found_auto_install" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Set to false to disable the "command not found" handler to autoinstall missing tool versions."###,
    },
    "os" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"OS to use for precompiled binaries."###,
    },
    "override_config_filenames" => SettingsMeta {
        type_: SettingsType::ListString,
        description: r###"If set, mise will ignore default config files like `mise.toml` and use these filenames instead. This must be an env var."###,
    },
    "override_tool_versions_filenames" => SettingsMeta {
        type_: SettingsType::ListString,
        description: r###"If set, mise will ignore .tool-versions files and use these filenames instead. Can be set to `none` to disable .tool-versions. This must be an env var."###,
    },
    "paranoid" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Enables extra-secure behavior."###,
    },
    "pin" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Default to pinning versions when running `mise use` in mise.toml files."###,
    },
    "pipx_uvx" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use uvx instead of pipx if uv is installed and on PATH."###,
    },
    "plugin_autoupdate_last_check_duration" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"How long to wait before updating plugins automatically (note this isn't currently implemented)."###,
    },
    "profile" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Profile to use for mise.${MISE_PROFILE}.toml files."###,
    },
    "python_compile" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"If true, compile python from source. If false, use precompiled binaries. If not set, use precompiled binaries if available."###,
    },
    "python_default_packages_file" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Path to a file containing default python packages to install when installing python."###,
    },
    "python_patch_url" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"URL to fetch python patches from."###,
    },
    "python_patches_directory" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Directory to fetch python patches from."###,
    },
    "python_precompiled_arch" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Specify the architecture to use for precompiled binaries."###,
    },
    "python_precompiled_os" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Specify the OS to use for precompiled binaries."###,
    },
    "python_pyenv_repo" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"URL to fetch pyenv from for compiling python."###,
    },
    "python_venv_auto_create" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Automatically create virtualenvs for python tools."###,
    },
    "python_venv_stdlib" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Prefer to use venv from Python's standard library."###,
    },
    "quiet" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Suppress all output except errors."###,
    },
    "raw" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Connect stdin/stdout/stderr to child processes."###,
    },
    "shorthands_file" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Path to a file containing custom tool shorthands."###,
    },
    "silent" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Suppress all `mise run|watch` output except errors—including what tasks output."###,
    },
    "system_config_file" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Path to the system mise config file. Default is `/etc/mise/config.toml`. This must be an env var."###,
    },
    "task_disable_paths" => SettingsMeta {
        type_: SettingsType::ListPath,
        description: r###"Paths that mise will not look for tasks in."###,
    },
    "task_output" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Change output style when executing tasks."###,
    },
    "task_remote_no_cache" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Mise will always fetch the latest tasks from the remote, by default the cache is used."###,
    },
    "task_run_auto_install" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Automatically install missing tools when executing tasks."###,
    },
    "task_skip" => SettingsMeta {
        type_: SettingsType::SetString,
        description: r###"Tasks to skip when running `mise run`."###,
    },
    "task_timings" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Show completion message with elapsed time for each task on `mise run`. Default shows when output type is `prefix`."###,
    },
    "trace" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Sets log level to trace"###,
    },
    "trusted_config_paths" => SettingsMeta {
        type_: SettingsType::ListPath,
        description: r###"This is a list of config paths that mise will automatically mark as trusted."###,
    },
    "unix_default_file_shell_args" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"List of default shell arguments for unix to be used with `file`. For example `sh`."###,
    },
    "unix_default_inline_shell_args" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"List of default shell arguments for unix to be used with inline commands. For example, `sh`, `-c` for sh."###,
    },
    "use_file_shell_for_executable_tasks" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Determines whether to use a specified shell for executing tasks in the tasks directory. When set to true, the shell defined in the file will be used, or the default shell specified by `windows_default_file_shell_args` or `unix_default_file_shell_args` will be applied. If set to false, tasks will be executed directly as programs."###,
    },
    "use_versions_host" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Set to false to disable using the mise-versions API as a quick way for mise to query for new versions."###,
    },
    "verbose" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Shows more verbose output such as installation logs when installing tools."###,
    },
    "vfox" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use vfox as a default plugin backend instead of asdf."###,
    },
    "windows_default_file_shell_args" => SettingsMeta {
        type_: SettingsType::ListString,
        description: r###"List of default shell arguments for Windows to be used for file commands. For example, `cmd`, `/c` for cmd.exe."###,
    },
    "windows_default_inline_shell_args" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"List of default shell arguments for Windows to be used for inline commands. For example, `cmd`, `/c` for cmd.exe."###,
    },
    "windows_executable_extensions" => SettingsMeta {
        type_: SettingsType::ListString,
        description: r###"List of executable extensions for Windows. For example, `exe` for .exe files, `bat` for .bat files, and so on."###,
    },
    "windows_shim_mode" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Shim file mode for Windows. Options: `file`, `hardlink`, `symlink`."###,
    },
    "yes" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"This will automatically answer yes or no to prompts. This is useful for scripting."###,
    },
    "aqua.baked_registry" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use baked-in aqua registry (if compiled in)."###,
    },
    "aqua.cosign" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use cosign to verify aqua tool signatures."###,
    },
    "aqua.cosign_extra_args" => SettingsMeta {
        type_: SettingsType::ListString,
        description: r###"Extra arguments to pass to cosign when verifying aqua tool signatures."###,
    },
    "aqua.minisign" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use minisign to verify aqua tool signatures."###,
    },
    "aqua.registry_url" => SettingsMeta {
        type_: SettingsType::Url,
        description: r###"URL to fetch aqua registry from."###,
    },
    "aqua.slsa" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use SLSA to verify aqua tool signatures."###,
    },
    "cargo.binstall" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use cargo-binstall instead of cargo install if available"###,
    },
    "cargo.registry_name" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Name of the cargo registry to use."###,
    },
    "dotnet.package_flags" => SettingsMeta {
        type_: SettingsType::ListString,
        description: r###"Extends dotnet search and install abilities."###,
    },
    "dotnet.registry_url" => SettingsMeta {
        type_: SettingsType::Url,
        description: r###"URL to fetch dotnet tools from."###,
    },
    "erlang.compile" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"If true, compile erlang from source. If false, use precompiled binaries. If not set, use precompiled binaries if available."###,
    },
    "node.compile" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Compile node from source."###,
    },
    "node.flavor" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Install a specific node flavor like glibc-217 or musl. Use with unofficial node build repo."###,
    },
    "node.gpg_verify" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use gpg to verify node tool signatures."###,
    },
    "node.mirror_url" => SettingsMeta {
        type_: SettingsType::Url,
        description: r###"Mirror to download node tarballs from."###,
    },
    "npm.bun" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use bun instead of npm if bun is installed and on PATH."###,
    },
    "pipx.registry_url" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"URL to use for pipx registry."###,
    },
    "pipx.uvx" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use uvx instead of pipx if uv is installed and on PATH."###,
    },
    "python.compile" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"If true, compile python from source. If false, use precompiled binaries. If not set, use precompiled binaries if available."###,
    },
    "python.default_packages_file" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Path to a file containing default python packages to install when installing a python version."###,
    },
    "python.patch_url" => SettingsMeta {
        type_: SettingsType::Url,
        description: r###"URL to fetch python patches from to pass to python-build."###,
    },
    "python.patches_directory" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Directory to fetch python patches from."###,
    },
    "python.precompiled_arch" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Specify the architecture to use for precompiled binaries."###,
    },
    "python.precompiled_flavor" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Specify the flavor to use for precompiled binaries."###,
    },
    "python.precompiled_os" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Specify the OS to use for precompiled binaries."###,
    },
    "python.pyenv_repo" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"URL to fetch pyenv from for compiling python with python-build."###,
    },
    "python.uv_venv_auto" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Integrate with uv to automatically create/source venvs if uv.lock is present."###,
    },
    "python.uv_venv_create_args" => SettingsMeta {
        type_: SettingsType::ListString,
        description: r###"Arguments to pass to uv when creating a venv."###,
    },
    "python.venv_auto_create" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Automatically create virtualenvs for python tools."###,
    },
    "python.venv_create_args" => SettingsMeta {
        type_: SettingsType::ListString,
        description: r###"Arguments to pass to python when creating a venv. (not used for uv venv creation)"###,
    },
    "python.venv_stdlib" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Prefer to use venv from Python's standard library."###,
    },
    "ruby.apply_patches" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"A list of patch files or URLs to apply to ruby source."###,
    },
    "ruby.default_packages_file" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Path to a file containing default ruby gems to install when installing ruby."###,
    },
    "ruby.ruby_build_opts" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Options to pass to ruby-build."###,
    },
    "ruby.ruby_build_repo" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"URL to fetch ruby-build from."###,
    },
    "ruby.ruby_install" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use ruby-install instead of ruby-build."###,
    },
    "ruby.ruby_install_opts" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Options to pass to ruby-install."###,
    },
    "ruby.ruby_install_repo" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"URL to fetch ruby-install from."###,
    },
    "ruby.verbose_install" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Set to true to enable verbose output during ruby installation."###,
    },
    "rust.cargo_home" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Path to the cargo home directory. Defaults to `~/.cargo` or `%USERPROFILE%\.cargo`"###,
    },
    "rust.rustup_home" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Path to the rustup home directory. Defaults to `~/.rustup` or `%USERPROFILE%\.rustup`"###,
    },
    "sops.age_key" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"The age private key to use for sops secret decryption."###,
    },
    "sops.age_key_file" => SettingsMeta {
        type_: SettingsType::Path,
        description: r###"Path to the age private key file to use for sops secret decryption."###,
    },
    "sops.age_recipients" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"The age public keys to use for sops secret encryption."###,
    },
    "sops.rops" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use rops to decrypt sops files. Disable to shell out to `sops` which will slow down mise but sops may offer features not available in rops."###,
    },
    "status.missing_tools" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Show a warning if tools are not installed when entering a directory with a mise.toml file."###,
    },
    "status.show_env" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Show configured env vars when entering a directory with a mise.toml file."###,
    },
    "status.show_tools" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Show configured tools when entering a directory with a mise.toml file."###,
    },
    "status.truncate" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Truncate status messages."###,
    },
    "swift.gpg_verify" => SettingsMeta {
        type_: SettingsType::Bool,
        description: r###"Use gpg to verify swift tool signatures."###,
    },
    "swift.platform" => SettingsMeta {
        type_: SettingsType::String,
        description: r###"Override the platform to use for precompiled binaries."###,
    },
    }
});
    